# encoding: utf-8
# Created by chenghaomou at 2019-05-01
import re
from dataclasses import dataclass
from itertools import product
from Patcher.Utils import Romanizer, debug


@dataclass
class Retriever:

    candidates: dict
    language: str
    romanizer: Romanizer
    lowercase: bool = True
    top_k: int = 5
    vocab: list = None
    rom_vocab: list = None
    rom2norm: dict = None
    inited: bool = False

    def post_init(self) -> None:
        if not self.inited:
            self.vocab = list(self.candidates.keys())
            self.rom_vocab = self.romanizer.run(' ##### '.join(self.vocab).lower() if self.lowercase else ' ##### '.join(self.vocab)).split(' ##### ')
            self.rom2norm = {r: n for r, n in zip(self.rom_vocab, self.vocab)}
            self.inited = True
        return

    def search(self, token: str) -> list:
        ranking = self.most_similar(token)

        if ranking:
            return ranking[0][1]
        else:
            return [token]

    def search_online(self, token: str, translate_client: callable = None) -> list:
        from google.cloud import translate
        if translate_client is None:
            translate_client = translate.Client()
        translation = translate_client.translate(token, target_language='en', source_language=self.language)

        return [translation['translatedText']]

    @debug
    def most_similar(self, token: str) -> list:

        from weighted_levenshtein import dam_lev
        import numpy as np

        if token.isupper(): return [(token, [token])]
        if token[0].isupper(): return [(token, [token])]
        if any(c.isupper() for c in token):
            res = re.findall(r'([A-Z][a-z]+)', token)
            return [(token, res)] if res else [(token, [token])]

        self.post_init()

        insert_costs = np.ones(128, dtype=np.float64)
        delete_costs = np.ones(128, dtype=np.float64)
        substitute_costs = np.ones((128, 128), dtype=np.float64)
        transpose_costs = np.ones((128, 128), dtype=np.float64)

        token = self.romanizer.run(token)
        token = token.lower() if self.lowercase else token

        for c in set(token):
            insert_costs[ord(c)] = 0.9
            delete_costs[ord(c)] = 0.9

        for c1, c2 in product(set(token), set(token)):
            if c1 != c2:
                substitute_costs[ord(c1), ord(c2)] = 0.9
                substitute_costs[ord(c2), ord(c1)] = 0.9

                transpose_costs[ord(c1), ord(c2)] = 0.9
                transpose_costs[ord(c2), ord(c1)] = 0.9

        score = lambda x: dam_lev(x,
                                  token,
                                  transpose_costs=transpose_costs,
                                  insert_costs=insert_costs,
                                  delete_costs=delete_costs,
                                  substitute_costs=substitute_costs
                                  )

        ranking = sorted(self.rom_vocab, key=score)

        ranking = [(self.rom2norm[r], self.candidates[self.rom2norm[r]]) for r in ranking[:self.top_k]]

        return ranking








